/* Copyright (C) 1999 PLUS Group - University of Maryland
 *	
 * This software is provided under the terms of the GNU General Public
 * License Version 2, June 1991. See the header in KA_App.java
 * and the file License.html for details.
 */

/* Change History:
    11/03/00, JDH: Used accessor methods for fields in Instance
              JDH: Use new constructor for Instance
*/

package Annotator;

import java.awt.*;
import java.util.*;
import java.io.*;
import java.net.*;
import UserInt.*;
import SHOE.*;
import HTML.*;
import java.lang.System;

public class InstWindow extends EditWindow {

  public static final int ADD = 1;
  public static final int EDIT = 2;

  int winMode;
  SHOE_Doc edDoc;
  Instance edInst = null;
  ObjectList instList;
  Label l_key, l_del;
  TextField keyText, delText;

  Label  l_name;
  TextField  nameText;
  Button  nameButton;

  Relation relName=null;
  OntManager ontman;
  String strPageAddress;

  int edIndex = 0;

  public InstWindow(Frame parent, int mode, SHOE_Doc doc, ObjectList list,
		    OntManager om, String address) {
    
    super(parent, "Instance");

    winMode = mode;
    instList = list;
    edDoc = doc;
    ontman=om;
    strPageAddress=address;

    Panel textPanel = new Panel();
    textPanel.setLayout(new GridBagLayout());
    l_key = new Label("Key:");
    keyText = new TextField(40);
    keyText.setBackground(Color.white);
    l_del = new Label("Delegate-To:");
    delText = new TextField(40);
    delText.setBackground(Color.white);

    Utilities.placeComponent(textPanel, l_key, 0, 0, 1, 1,
			     GridBagConstraints.EAST);
    Utilities.placeComponent(textPanel, keyText, 1, 0, 3, 1);
    Utilities.placeComponent(textPanel, l_del, 0, 1, 1, 1,
			     GridBagConstraints.EAST);
    Utilities.placeComponent(textPanel, delText, 1, 1, 3, 1);
    
    ///////////// 7/11/98 /////////////
    l_name = new Label("Name:");
    nameText = new TextField(40);
    nameText.setBackground(Color.white);
    nameButton= new Button("Using Title");

    Utilities.placeComponent(textPanel, l_name, 0, 2, 1, 1,
			     GridBagConstraints.EAST);
    Utilities.placeComponent(textPanel, nameText, 1, 2, 2, 1);
    Utilities.placeComponent(textPanel, nameButton, 3, 2, 1, 1);
    //////////////////////////////////	
    add("North", textPanel);

    if (winMode == ADD) {
      setTitle("Add Instance");
	keyText.setText(strPageAddress);
    }
    else {
      setTitle("Edit Instance");
      edIndex = instList.getSelectedIndex();
      edInst = (Instance)instList.getSelectedObject();
      // System.out.println ("Index = " + edIndex + ", Object = " + edInst);
      keyText.setText(edInst.getKey());
      delText.setText(edInst.getDelegateStr());
	nameText.setText(getName(edInst));
    }
    pack();
    resize(minimumSize());
  }

  /* Searches the instances declarations for a relation named "name" and
     returns the first one whose first argument matches the instance's key
     (i.e., a name relation whose subject is the instance) */
  String getName(Instance inst) {
     SHOE_Obj obj;
     String arg,name=null;	
     Enumeration claim_enum= inst.getDeclarations();
	
     while (claim_enum.hasMoreElements()) {
	 obj = (SHOE_Obj)claim_enum.nextElement();
	 if(obj instanceof Relation) {
	   Relation rel = (Relation)obj;
	   arg=rel.getArg(1).getValue();
	   if(arg.equals(inst.getKey()) && rel.getBaseName().equals("name")) {
		relName=rel; 
	      name=(String)rel.getArg(2).getValue();	
		break;
	   }
	 }
     }
     return name;
  }


  public void saveChanges() {

    StringTokenizer tknzr;
    Vector delVector;
    String[] delArray;
    String delStr;

    delVector = new Vector();
    delStr = delText.getText();
    if (delStr != null) {
      tknzr = new StringTokenizer(delStr);
      while (tknzr.hasMoreTokens())
        delVector.addElement(tknzr.nextToken());
    }
    delArray = new String[delVector.size()];
    delVector.copyInto(delArray);
    if (winMode == ADD) {
      edInst = new Instance(keyText.getText(), delArray);
      instList.addItem(edInst.getKey(), edInst);
      edDoc.addInstance(edInst);

	String name;
	name=nameText.getText();
	if(!name.equals("")) {
	  Relation rel = new Relation(".name");
	  rel.addArg(new RelArg(1, keyText.getText()));
	  rel.addArg(new RelArg(2, name));
	  edInst.addDeclaration(rel, ontman);
	}
    }
    else {
      edInst.update(keyText.getText(), delVector);
      instList.replaceItem(edInst.getKey(), edInst, edIndex);
      
      String name;
      name=nameText.getText();
      if(!name.equals("")) {
	if(relName==null) {
	  Relation rel = new Relation(".name");
	  rel.addArg(new RelArg(1, keyText.getText()));
	  rel.addArg(new RelArg(2, name));
	  edInst.addDeclaration(rel, ontman);
	}
	else {
          relName.setArgValue(1, keyText.getText());
	  relName.setArgValue(2, name);
	  relName.validate(edInst,ontman);
	}
      }
      else 
	if(relName != null) edInst.deleteDeclaration(relName);
    }
  }

  public boolean action(Event event, Object arg){
    super.action(event, arg);
    if (event.target instanceof Button) {
      if(event.target == nameButton){
	nameText.setText(getPageTitle(keyText.getText()));
	return false;
      }
    }
    return false;
  }

  /* Gets the title of the page identified by the value entered in the key
     field */
  String getPageTitle(String key) {

    SHOE_Doc instDoc = null;
    AltProxyUrl url;
    InputStream inStream;
    Enumeration enum;
    HTML_Token token;
    SGML_Tag sgmltag;
    String title="";

    if (key != null) {
      if (key.equals(strPageAddress))   /* key is URL of current document */
	instDoc = edDoc;
      else {                     /* parse document specified by the URL */
	instDoc = new SHOE_Doc();
	try {
	  /* we use AltProxyUrl so when this is called from within an
	     applet, it will use the proxy server to fetch web pages from
	     other servers, and thus avoid Java applet security
	     restrictions. */
	  url = new AltProxyUrl(key);
	  inStream = url.openStream();
	  instDoc.parse(inStream);
	} catch (MalformedURLException ex) {
	  return title;
	} catch (IOException ex) {
	  return title;
	}

      }
    }

    enum = instDoc.getHTML();
    while (enum.hasMoreElements()) {
      token = (HTML_Token)enum.nextElement();
      if (token instanceof SGML_Tag) {
        sgmltag = (SGML_Tag)token;
        if (sgmltag.getName().equals("TITLE")) {
	  token = (HTML_Token)enum.nextElement();
	  title=token.toString();
	  break;	  
	}
      }
    }
	
    return title;
  }



}


