/* Knowledge Annotator
 * Copyright (C) 1999 PLUS Group - University of Maryland
 *	
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * Authors: Jeff Heflin and Qin Zhendong
 * Contact: Jeff Heflin (heflin@cs.umd.edu)
 * 
 */

/* Change history:
   01/29/01, JDH: Use new version of Instance.getUseOnts()
   11/03/00, JDH: Used accessor methods for fields in Instance
             JDH: Use new constructor for Instance
   10/26/00, JDH: Used accessor methods for fields in Category, Relation,
                  and Use_Ontology
   */

package Annotator;

import java.awt.*;
import java.io.*;
import java.util.*;
import java.net.*;
import UserInt.*;
import SHOE.*;
import Annotator.*;
import HTML.AltProxyUrl;

public class KA_App extends Frame {

  public static final String versionNum = "0.93";

  protected static int DEFAULT_WIDTH  = 500;
  protected static int DEFAULT_HEIGHT = 400;

  /* use the following to pass current URL to instance window so that
     the key can default to the current URL. We may be able to get rid
     of this */
  public String pageURL="";

  SHOE_Doc doc = new SHOE_Doc();
  OntManager ontman;                          // used to cache ontologies
  File SHOE_File = null;
  Thread dlgThrd = null;                     // used to process dialog boxes

  ArrayMenuBar menuBar;
  CheckboxMenuItemGroup viewMenuGroup;
  ObjectList instList, useOntList, claimList;
  Button b_AddInst, b_EditInst, b_DelInst;
  Button b_AddUseOnt, b_EditUseOnt, b_DelUseOnt;
  Button b_AddClaim, b_EditClaim, b_DelClaim;
  FileDialog openFileDialog, saveFileDialog;
  Panel viewPanel;
  TextArea sourceText;
  TextField statusBar;

  ///////// ClipBoard ---- 5/29/98, added by Qin ///////
  static Vector clipClaimVector;
  static Vector clipUseOntVector;
  static Vector clipInstanceVector;

  static Vector vSubjectModules=new Vector(); 
  static Vector vSubjectTexts=new Vector(); 
  static Vector vSubjectUseOnts=new Vector(); 

  static Menu mBookmarks; //=new Menu("Bookmarks");
  static Menu mAddBK=new Menu("Add Bookmark");

  static final String PROXY_FILE = "ontproxy.dat";

  /* view constants */
  static final int V_PHRASE = 1;         // claims as phrases
  static final int V_PREDICATE = 2;      // claims as predicates
  static final int V_SUMMARY = 3;        // summary of all SHOE
  static final int V_SOURCE = 4;         // source HTML

  int currentView = V_PHRASE;            // view that is visible
  boolean sourceEdit = false;            // edits made in source view?

  boolean useCheckboxWorkAround = false;

  public static void main(String args[]) {

    KA_App app = new KA_App();
    ApplType.setApplication();  // if called from main, must be an application
    // app.resize(500,400);
    // app.show();
    // app.repaint();

    /* since this is an application, enable the "Open File" and
       "Edit Bookmarks" menus */
    app.menuBar.getMenu("File").getItem("Open File").enable(); 
    app.menuBar.getMenu("Bookmarks").enable(); 
    app.menuBar.getMenu("Bookmarks").getItem("Edit Bookmarks").enable();      

    /* allow the use of bookmarks if running as an application. We don't
       allow it for the Applet because right now the bookmarks file uses
       object serialization, a feature of Java 1.1 */
    app.loadBookmarks();

    app.loadOntProxies();
  }

  public KA_App() {

    super("Knowledge Annotator");
    setup();
    show();
    ontman = new OntManager();
    ApplType.printSystemProperties();
    ApplType.setApplet();       // assume it is an Applet
                                // this must be over-ridden by applications
    setCheckboxWorkAroundStatus();
  }

  void setup() {

    clipClaimVector = new Vector();
    clipUseOntVector = new Vector();
    clipInstanceVector = new Vector();

    // setFont(new Font("TimesRoman", Font.PLAIN, 12));
    setFont(new Font("Helvetica", Font.PLAIN, 12));
    setBackground(Color.lightGray);
    setupMenuBar();
    setupPanels();
  }


  /* ***************** Bookmark functions **************************** */

  void loadBookmarks() {
    int i;
	
    for (i=0; i< menuBar.countMenus(); i++) {
      /* Java 1.1 version is below */
    // for(i=0;i<menuBar.getMenuCount(); i++) {
	mBookmarks=menuBar.getMenu(i);
	if(mBookmarks.getLabel().equals("Bookmarks")) break;
    }
    mBookmarks.addSeparator(); 

    try {
      FileInputStream istream = new FileInputStream("Bookmarks.dat");
      ObjectInputStream in = new ObjectInputStream(istream);

      int countBK=in.readInt();
      for(i=0;i<countBK; i++) {	
	MenuItem m;
	String label=in.readUTF();
	String type=in.readUTF();
	if(type.equals("Item")) { 
	  String obj=in.readUTF();
	  m=new ObjectMenuItem(label,obj);
	}
	else {
	  m=new Menu(label);
	  readFolder((Menu)m,in);
	}
	mBookmarks.add(m);
      }      
      in.close();
    } catch(IOException  e) {
      /* don't want to stop user for this message */
      statusBar.setText("Failed to load Bookmarks!");
    }
  }


  void readFolder(Menu menu,ObjectInputStream in) throws IOException{
	int i;

	int countMenu=in.readInt();
      for(i=0;i<countMenu; i++) {	
	  MenuItem m;
	  String label=in.readUTF();
	  String type=in.readUTF();
	  if(type.equals("Item")) { 
		String obj=in.readUTF();
		m=new ObjectMenuItem(label,obj);
	  }
	  else {
		m=new Menu(label);	  
		readFolder((Menu)m,in);
	  }
        menu.add(m);
	}
  }

  void writeFolder(Menu menu,ObjectOutputStream out) throws IOException{
    int i;

    // out.writeInt(menu.getItemCount()); 
    // for(i=0;i<menu.getItemCount(); i++) {	
    out.writeInt(menu.countItems()); 
    for(i=0;i<menu.countItems(); i++) {	
      MenuItem m=menu.getItem(i);
      out.writeUTF(m.getLabel());
      if(m instanceof ObjectMenuItem) {
	out.writeUTF("Item");
	out.writeUTF((String)((ObjectMenuItem)m).getObject());
      }
      else {
	out.writeUTF("Folder");
	writeFolder((Menu)m,out);
      }
    }
  }

  void saveBookmarks() {
    int i;

    try {
      FileOutputStream ostream = new FileOutputStream("Bookmarks.dat");
      ObjectOutputStream out = new ObjectOutputStream(ostream);

      // out.writeInt(mBookmarks.getItemCount() - 3); 
      // for(i=3; i<mBookmarks.getItemCount(); i++) {	
      out.writeInt(mBookmarks.countItems() - 3); 
      for(i=3; i<mBookmarks.countItems(); i++) {	
	MenuItem m=mBookmarks.getItem(i);
	out.writeUTF(m.getLabel());
	if(m instanceof ObjectMenuItem) {
	  out.writeUTF("Item");
	  out.writeUTF((String)((ObjectMenuItem)m).getObject());
	}
	else {
	  out.writeUTF("Folder");
	  writeFolder((Menu)m,out);
	}
      }

      out.close();
    } catch(IOException  e) {
      notifyUser("Failed to save Bookmarks!");
      return;
    }
  }

  /* ************************************************************* */

  /** Loads the ontology proxy file. Each line specifies an ontology id,
    version and a proxy URL for retrieving it. These proxies are then
    set in the OntManager.
    */
  void loadOntProxies() {

    FileInputStream inStream;
    String errMsg;

    File proxyFile = new File(PROXY_FILE);    
    try {
      inStream = new FileInputStream(proxyFile);
      errMsg = ontman.loadOntProxyInputStream(inStream);
      if (errMsg != null) {
	notifyUser("Found errors while reading ontology proxy file " + 
		   PROXY_FILE + ":\n" + errMsg);
      }
    } catch (FileNotFoundException e) {
      statusBar.setText("There is no ontology proxy file: " + PROXY_FILE);
    }
  }

  void setupMenuBar() {
    Object menuItems[][] = {
      {"File", "Open URL", "~Open File", "-", "~Save File", "~Save File As",
       "-", "Validate", "-", "Exit"},
      {"View","+Claim Phrases", "-Claim Predicates", "-Source HTML", "-Summary"},
      {"~Instance", "Add", "~Edit", "~Delete","-","Copy","~Paste","~Clear"},
      {"~Use Ontology", "Add", "~Edit", "~Delete","-","Copy","~Paste","~Clear"},
      // {"~Claim", "Add Relation", "Add Category", "-", "~Edit", "~Delete"},
      {"~Claim", "Add", "~Edit", "~Delete","-","Copy","~Paste","~Clear"},
      /* not using subject module right now */
      // {"Subject","Add Subject Module","-"},
      {"~Bookmarks", "~Add Bookmark", "~Edit Bookmarks"},
      {"Help", "About Knowledge Annotator"},
    };
    menuBar = new ArrayMenuBar(menuItems);
    setMenuBar(menuBar);

    CheckboxMenuItem cmi;
    viewMenuGroup = new CheckboxMenuItemGroup();
    cmi = (CheckboxMenuItem)menuBar.getMenu("View").getItem("Claim Phrases");
    viewMenuGroup.add(cmi);
    cmi = (CheckboxMenuItem)menuBar.getMenu("View").getItem("Claim Predicates");
    viewMenuGroup.add(cmi);
    cmi = (CheckboxMenuItem)menuBar.getMenu("View").getItem("Source HTML");
    viewMenuGroup.add(cmi);
    cmi = (CheckboxMenuItem)menuBar.getMenu("View").getItem("Summary");
    viewMenuGroup.add(cmi);
  }

  void setupPanels() {

    /* create a split panel for the GUI (Claim Phrases and Claim Predicates) */
    Panel guiPanel = new Panel();
    guiPanel.setLayout(new GridBagLayout());

    Utilities.placeComponent(guiPanel, new Label("Instances:"), 0, 0, 1, 1);
    instList = new ObjectList(5, false);
    Utilities.placeComponent(guiPanel, instList, 0, 1, 1, 5, 0.5, 0.5, true);
    Panel instButtons = new Panel();
    b_AddInst = new Button("Add");
    b_EditInst = new Button("Edit");
    b_DelInst = new Button("Delete");
    instButtons.add(b_AddInst);
    instButtons.add(b_EditInst);
    instButtons.add(b_DelInst);
    b_AddInst.disable();
    b_EditInst.disable();
    b_DelInst.disable();
    Utilities.placeComponent(guiPanel, instButtons, 0, 6, 1, 2);

    Utilities.placeComponent(guiPanel, new Label("Use Ontologies:"), 1, 0, 1, 1);
    useOntList = new ObjectList(5, false);
    Utilities.placeComponent(guiPanel, useOntList, 1, 1, 1, 5, 0.5, 0.5, true);
    Panel ontButtons = new Panel();
    b_AddUseOnt = new Button("Add");
    b_EditUseOnt = new Button("Edit");
    b_DelUseOnt = new Button("Delete");
    ontButtons.add(b_AddUseOnt);
    ontButtons.add(b_EditUseOnt);
    ontButtons.add(b_DelUseOnt);
    b_AddUseOnt.disable();
    b_EditUseOnt.disable();
    b_DelUseOnt.disable();
    Utilities.placeComponent(guiPanel, ontButtons, 1, 6, 1, 2);

    Utilities.placeComponent(guiPanel, new Label("Claims:"), 0, 8, 2, 1);
    claimList =  new ObjectList(10, false);
    Utilities.placeComponent(guiPanel, claimList, 0, 9, 2, 10, 1.0, 1.0, true);
    Panel claimButtons = new Panel();
    b_AddClaim = new Button("Add");
    b_EditClaim = new Button("Edit");
    b_DelClaim = new Button("Delete");
    claimButtons.add(b_AddClaim);
    claimButtons.add(b_EditClaim);
    claimButtons.add(b_DelClaim);
    b_AddClaim.disable();
    b_EditClaim.disable();
    b_DelClaim.disable();
    Utilities.placeComponent(guiPanel, claimButtons, 0, 19, 2, 1);

    /* Create a panel for the text area (Source HTML and Summary views) */
    Panel sourcePanel = new Panel();
    sourceText = new TextArea(20,80);
    sourceText.setBackground(Color.lightGray);
    sourceText.setEditable(false);
    sourcePanel.setLayout(new BorderLayout());
    sourcePanel.add("Center", sourceText);

    /* use a card panel to switch between the two views created above */
    viewPanel = new Panel();
    viewPanel.setLayout(new CardLayout());
    viewPanel.add("SHOE", guiPanel);
    viewPanel.add("Source", sourcePanel);
    ((CardLayout)viewPanel.getLayout()).show(viewPanel, "SHOE");

    /* create a main panel to contain the status bar and view panel */
    Panel mainPanel = new Panel();
    mainPanel.setLayout(new BorderLayout());
    mainPanel.add("Center", viewPanel);
    statusBar = new TextField();
    mainPanel.add("South", statusBar);

    // Note that the following does nothing in Java 1.2+. Due to a bug you
    // cannot change the color of an uneditable TextField! It is
    // always the color SystemColor.control!
    statusBar.setBackground(Color.lightGray);
    statusBar.setEditable(false); 

    add("Center", mainPanel);
    pack();
    // the following resize makes window super small in JDK 1.3
    // resize(new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT));
    resize(preferredSize());
    // For Java 1.1: 
    // resize(getPreferredSize());
  }

  public boolean handleEvent(Event event) {

    if (event.id == Event.WINDOW_DESTROY) {
      if (ApplType.getIsApplication()) {
	// saveSubjectModules();
	saveBookmarks();
	System.exit(0);
      }
      else
	dispose();
      return true;
    }

    else if (event.target instanceof ObjectList) {
      ObjectList list = (ObjectList)(event.target);
      /* Instance List event */
      if (list == instList) {
	if (event.id == Event.LIST_SELECT) {

	  Use_Ontology[] useOnts;
	  Enumeration enum;
	  Instance inst;

	  inst = (Instance)list.getSelectedObject();
	  disableUseOntOps();
	  disableInstOps();
	  useOnts = inst.getUseOnts();
	  replaceUseOntList(useOnts);
	  enum = inst.getDeclarations();
	  replaceClaimList(enum);
	  enableInstOps();
	}
      }
      else if (list == useOntList) {
	if (event.id == Event.LIST_SELECT) {
	  enableUseOntOps();
	}
      }
      else if (list == claimList) {
	if (event.id == Event.LIST_SELECT) {
	  enableClaimOps();
	}
      }
    }
    else if (event.target == sourceText) {
      if (event.id == Event.KEY_PRESS) {     // text edited in the source view
	sourceEdit = true;
      }
    }

    return super.handleEvent(event);
  }

  public boolean action(Event e, Object arg) {

    Object target = e.target;

    if (target instanceof MenuItem) {
      MenuItem menu = (MenuItem)e.target; 
      MenuItem pmenu = (MenuItem)menu.getParent();
      MenuComponent  ppmenu = pmenu;
      MenuComponent  pppmenu= (MenuComponent)ppmenu.getParent();

      while(!(pppmenu instanceof MenuBar)) {
	ppmenu=pppmenu;
	pppmenu=(MenuComponent)ppmenu.getParent();
      }
      if (pmenu.getLabel().equals("File")) {
	if ("Open File".equals(e.arg)) 
	  processOpenFile();
	else if ("Open URL".equals(e.arg)) {
	  dlgThrd = new OpenURLThread(this);
	  dlgThrd.start();
	  // System.out.println("OpenURLThread started");
	}
	else if ("Save File".equals(e.arg))
	  processSaveFile();
	else if ("Save File As".equals(e.arg))
	  processSaveAsFile();
	else if ("Validate".equals(e.arg))
	  processValidate();
	else if ("Exit".equals(e.arg)) {
	  if (ApplType.getIsApplication()) {
	    // saveSubjectModules();
	    saveBookmarks();
	    System.exit(0);
	  }
	  else
	    dispose();
	}
      }

      /* Note: CheckboxMenuItem Events
	 Event handling for CheckboxMenuItems is different on different
	 platforms. Here's the analysis:
	 platforms:
	 1) JDK - Sun sparcs:  Two events are fired for the same target,
	    one has arg set to the state of the checkbox, the other
	    has arg set to the name of the menu item
	    (thus using menu.getLabel() would recognize and fire on 
	    both events)
         2) Netscape - Sun, Windows & Mac: One event is fired, arg is set to
	    the name of the menu item
	 3) IE - Windows: One event is fired, arg is set to the boolean
	    state of the checkbox
	 4) IE - Mac (java.version="1.1.7", java.vendor="Apple Computer Inc."):
	    Two identical events are fired, both have arg set to the name
	    of the menu item
	    */
      else if (pmenu.getLabel().equals("View")) {
	// System.out.println("View menu event fired: " + e);

	/* In all cases, the target is menu item object itself, and 
	   getLabel() will give us the name of the menu item. However,
	   since Solaris fires two event, we set the work-around variable
	   for it, and use the arg variable to check which item was
	   selected. */
	/* Note, we still need to fix the problem for IE on a Mac */
	String submenu;
	if (useCheckboxWorkAround == true)
	  submenu = e.arg.toString();
	else
	  submenu = menu.getLabel();

	if ("Claim Phrases".equals(submenu)) {
	  // System.out.println("Selected Claim Phrases");
	  // turn on Claim Phrases, turn off others
	  viewMenuGroup.unsetOthers((CheckboxMenuItem)menu);
	  processViewClaimPhrases();
	}
	else if ("Claim Predicates".equals(submenu)) {
	  // System.out.println("Selected Claim Predicates");
	  viewMenuGroup.unsetOthers((CheckboxMenuItem)menu);
	  processViewClaimPredicates();
	}
	else if ("Source HTML".equals(submenu)) {
	  // System.out.println("Selected Source");
	  // turn on source, turn off others
	  viewMenuGroup.unsetOthers((CheckboxMenuItem)menu);
	  processViewSource();
	}
	else if ("Summary".equals(submenu)) {
	  // System.out.println("Selected Summary");
	  // turn on summary, turn off others
	  viewMenuGroup.unsetOthers((CheckboxMenuItem)menu);
	  processViewSummary();
	}
	else {
	  // System.err.println("Could not match event to menu label " + submenu);
	}
      }

      else if (pmenu.getLabel().equals("Instance")) {
	if ("Add".equals(e.arg)) {
	  dlgThrd = new AddInstanceThread(this);
	  dlgThrd.start();
	  // System.out.println("AddInstanceThread started");
	}
	else if ("Edit".equals(e.arg)) {
	  dlgThrd = new EditInstanceThread(this);
	  dlgThrd.start();
	}
	else if ("Delete".equals(e.arg))
	  processDelInst();	  
	else if("Copy".equals(menu.getLabel())) {
	  clipInstanceCopy();
	}
	else if("Paste".equals(menu.getLabel())) {
	  clipInstancePaste();
	}	  
	else if("Clear".equals(menu.getLabel())) {
	  clipInstanceClear();
	}
      }

      else if (pmenu.getLabel().equals("Use Ontology")) {
	if ("Add".equals(e.arg)) {
	  // processAddUseOnt();
	  dlgThrd = new AddUseOntThread(this);
	  dlgThrd.start();
	}
	else if ("Edit".equals(e.arg)) {
	  // processEditUseOnt();
	  dlgThrd = new EditUseOntThread(this);
	  dlgThrd.start();
	}
	else if ("Delete".equals(e.arg))
	  processDelUseOnt();
	else if("Copy".equals(menu.getLabel())) {
	  clipUseOntCopy();
	}
	else if("Paste".equals(menu.getLabel())) {
	  clipUseOntPaste();
	}	  
	else if("Clear".equals(menu.getLabel())) {
	  clipUseOntClear();
	}
      }

      else if (pmenu.getLabel().equals("Claim")) {
	if ("Add".equals(e.arg)) {
	  dlgThrd = new AddClaimThread(this);
	  dlgThrd.start();
	}
	// else if ("Add Relation".equals(e.arg))
	//  notifyUser("Add Relation is not available with this release");
	// else if ("Add Category".equals(e.arg))
	//  notifyUser("Add Category is not available with this release");
	else if ("Edit".equals(e.arg)) {
	  dlgThrd = new EditClaimThread(this);
	  dlgThrd.start();
	}
	else if ("Delete".equals(e.arg))
	  processDelClaim();
	else if("Copy".equals(menu.getLabel())) {
	  // System.out.println("Action: Claim-Copy");
	  clipClaimCopy();
	}
	else if("Paste".equals(menu.getLabel())) {
	  clipClaimPaste();
	}	  
	else if("Clear".equals(menu.getLabel())) {
	  clipClaimClear();
	}
      }

      else if(((Menu)ppmenu).getLabel().equals("Bookmarks")) {
	if("Add Bookmark".equals(e.arg)) {
	  ObjectMenuItem omi = new ObjectMenuItem(pageURL, pageURL);
	  mBookmarks.add(omi);
	}
	else if(menu.getLabel().equals("Edit Bookmarks")) {
	  notifyUser("This option is not yet implemented");
	}
	else if(menu instanceof ObjectMenuItem) {
	  ObjectMenuItem m = (ObjectMenuItem)menu;
	  //notifyUser((String)m.getObject());
	  loadPage((String)m.getObject());
	}
      }

      else if (pmenu.getLabel().equals("Help")) {
	if ("About Knowledge Annotator".equals(e.arg)) {
	  AboutWindow popup = new AboutWindow(this, versionNum);
	  // Utilities.showDialog(popup);
	  popup.showWorkaround();
	} 
      }

	/*
	////////////////////////////////////////////////////////////////////
	else if(((Menu)ppmenu).getLabel().equals("Subject")) {
	  int i,j;
	  Vector v;
	  for(i=0;i<vSubjectModules.size();i++) {
	    v= (Vector)vSubjectModules.elementAt(i);
	    if(((String)v.elementAt(0)).equals(pmenu.getLabel()))
	      break;
	  }
	  
	  if ("Add Subject Module".equals(event.arg))
	    addSubjectModule();
	  else if(menu.getLabel().equals("Delete"))
	    deleteSubjectModule(i); 
	  else if(menu.getLabel().equals("Paste as a new subject")) 
	    pasteSubjectModule(i);		   
	  else if(menu.getLabel().equals("Insert to current subject")) 
	    insertSubjectModule(i);		   
	  else {
	    // for(j=4;j<((Menu)pmenu).getItemCount();j++) {
	    for(j=4;j<((Menu)pmenu).countItems();j++) {
	      String str=menu.getLabel();	
	      if(str.equals(((Menu)pmenu).getItem(j).getLabel())){
		insertSubjectModule(i,j-4);		   
		break;
	      }
	    }
	  }
	}
	////////////////////////////////////////////////////////////////////
	*/

      return true;
    }

    if (target instanceof Button) {
      if (target == b_AddInst) {
      	// processAddInst();
	dlgThrd = new AddInstanceThread(this);
	dlgThrd.start();
	// System.out.println("Action thread: started AddInstanceThread");
      	return true;
      }
      else if (target == b_EditInst) {
	// processEditInst();
	dlgThrd = new EditInstanceThread(this);
	dlgThrd.start();
       	return true;
      }
      else if (target == b_DelInst) {
      	processDelInst();
      	return true;
      }
      else if (target == b_AddUseOnt) {
       	// processAddUseOnt();
	dlgThrd = new AddUseOntThread(this);
	dlgThrd.start();
      	return true;
      }
      else if (target == b_EditUseOnt) {
      	// processEditUseOnt();
	dlgThrd = new EditUseOntThread(this);
	dlgThrd.start();
       	return true;
      }
      else if (target == b_DelUseOnt) {
      	processDelUseOnt();
      	return true;
      }
      else if (target == b_AddClaim) {
       	// processAddClaim();
	dlgThrd = new AddClaimThread(this);
	dlgThrd.start();
      	return true;
      }
      else if (target == b_EditClaim) {
      	// processEditClaim();
	dlgThrd = new EditClaimThread(this);
	dlgThrd.start();
       	return true;
      }
      else if (target == b_DelClaim) {
      	processDelClaim();
      	return true;
      }
    }
    return false;
  }  

  /** Validates the SHOE in the current document and displays a dialog
    if with a list of errors if the document is invalid. */
  public void validateDoc(String location) {

    Dialog errWin;

    doc.validate(ontman);           // validate SHOE document
    if (doc.hasParseErrors()) {
      errWin = new ErrorDialog(this, "Parse Errors", 
               "The following parse errors were found in " + location + ":",
	       doc.getParseErrors());
      // Utilities.showDialog((WorkaroundDialog)errWin);
      ((WorkaroundDialog)errWin).showWorkaround();
    }
  }

  /** Updates the current view to display the current document. */
  public void displayDoc(SHOE_Doc doc) {

    Enumeration enum;

    /* no matter what view is current, set up the GUI stuff */
    enum = doc.getPSHOE();
    replaceInstList(enum);

    /*
    if (currentView == V_PHRASE)
      displayDocInPhraseView(doc);
    else if (currentView == V_PREDICATE)
      displayDocInPredView(doc);
      */
    if (currentView == V_SOURCE) 
      displayDocInSourceView(doc);
    else if (currentView == V_SUMMARY)
      displayDocInSummaryView(doc);
  }

  void displayDocInSourceView(SHOE_Doc doc) {

    ByteArrayOutputStream arrayStream;
    PrintStream htmlStream;

    arrayStream = new ByteArrayOutputStream();
    htmlStream = new PrintStream(arrayStream);
    doc.printHTML(htmlStream);
    sourceText.setText(arrayStream.toString());
    sourceText.setEditable(true);
    sourceEdit = false;
  }

  void displayDocInSummaryView(SHOE_Doc doc) {

    ByteArrayOutputStream arrayStream;
    PrintStream sumStream;

    arrayStream = new ByteArrayOutputStream();
    sumStream = new PrintStream(arrayStream);
    doc.printSummary(sumStream);
    sourceText.setText(arrayStream.toString());
    sourceText.setEditable(false);
  }

  public void loadPage(String str) {

    if(loadPageAsFile(str)==false)
      loadPageAsURL(str);
  }


  public boolean loadPageAsURL(String urlText) {

    AltProxyUrl url;
    InputStream inStream;
    Enumeration enum;
    
    SHOE_File = null;               // reset file object

    if (urlText != null) {
      doc.clear();                    // reset SHOE document object
      deselectInst();
	statusBar.setText("Reading URL: "+urlText);

      try {
      /* we use AltProxyUrl so when this is called from within an
	 applet, it will use the proxy server to fetch web pages from
	 other servers, and thus avoid Java applet security restrictions. */
        url = new AltProxyUrl(urlText);
        inStream = url.openStream();
        doc.parse(inStream);
        inStream.close();
		
	pageURL=urlText;
	/* 
	  if(mAddBK.getItemCount()!=0) mAddBK.remove(0);
	  ObjectMenuItem m = new ObjectMenuItem("Current Page",urlText);
	  mAddBK.insert(m,0);
	  */

      } catch (MalformedURLException ex) {
        //notifyUser("Error opening URL " + urlText);
	  statusBar.setText("Error: " + urlText + " is not a proper URL");
	  return false;
      } catch (IOException ex) {
        //notifyUser("Error reading URL " + urlText);
	  statusBar.setText("Error: Could not locate web page identified by " +
			    urlText);
        return false;
      } catch (Exception ex) {
	/* catch other exceptions that may be browser specific */
	notifyUser("Error opening URL: " + urlText + "\n" +
                   "   Message: " + ex.getMessage() + "\n" +
		   "   (Perhaps you have tried to open a web page that is " +
		   "not on the same server as the applet you are running)");
	return false;
      }
      
      validateDoc(urlText);
      displayDoc(doc);
      enableDocMenus();
      statusBar.setText("URL: "+urlText);
      // pack();
    }
    return true;
  }
    
  public boolean loadPageAsFile(String pageName) {

    FileInputStream inStream;
    Enumeration enum;

    if (pageName != null) {
      doc.clear();
      deselectInst();

      SHOE_File = new File(pageName);

      statusBar.setText("Reading file: "+SHOE_File);
	
      doc.clear();
      deselectInst();

      try {
	  inStream = new FileInputStream(SHOE_File);
      } catch (FileNotFoundException ex) {
	  statusBar.setText("Error opening file: "+SHOE_File);
 	  return false;
      }

      try {
	doc.parse(inStream);
	inStream.close();

	pageURL = pageName;
	/*
	  if(mAddBK.getItemCount()!=0) mAddBK.remove(0);
	  ObjectMenuItem m = new ObjectMenuItem("Current Page",SHOE_File.getPath() );
	  mAddBK.insert(m,0);
	  */

      } catch (IOException ex) {
	  statusBar.setText("Error reading file: "+SHOE_File);
	  return true;
      }

      validateDoc(SHOE_File.toString());
      displayDoc(doc);
      enableDocMenus();
      statusBar.setText("File: "+SHOE_File);
      // pack();
    }
    return true;
  }

  /** Parses a string to update the display with for viewing/editing
    a SHOE page represented by the string. */
  public void loadPageFromString(String pageSource) {

    byte  buf[];
    ByteArrayInputStream arrayStream;
    Enumeration enum;
    
    // create an input stream from the string
    buf = new byte[pageSource.length()];
    pageSource.getBytes(0, pageSource.length(), buf, 0);
    arrayStream = new ByteArrayInputStream(buf);

    doc.clear();                    // reset SHOE document object
    deselectInst();
    statusBar.setText("Reparsing document...");

    try {
      doc.parse(arrayStream);
    } catch (IOException ex) {
      statusBar.setText("Error parsing changes");
      return;
    }

    validateDoc("the document");
    displayDoc(doc);
    statusBar.setText("Parse complete.");
    // pack();
  }


  public void processOpenFile() {

    FileInputStream inStream;
    Enumeration enum;

//    SHOE_Doc doc = new SHOE_Doc();

    /* IE doesn't like applets that create file dialogs. Therefore we
       don't create the dialog until we need it */
    if (openFileDialog == null)
      openFileDialog = new FileDialog(this, "Open File", FileDialog.LOAD);
    openFileDialog.show();

    if (openFileDialog.getFile() != null) {
      SHOE_File = new File(openFileDialog.getDirectory(),
			   openFileDialog.getFile());
      // System.out.println("Opening " + SHOE_File.toString());
      loadPageAsFile(SHOE_File.toString());
    }
    else
      statusBar.setText("Open File canceled");
    return;
  }

  public void processOpenURL() {

    URLDialog openURLDialog = new URLDialog(this);
    String urlText;
    
    // Utilities.showDialog(openURLDialog);
    openURLDialog.showWorkaround();
    urlText = openURLDialog.getURL();

    if (urlText != null)
      loadPageAsURL(urlText);
    else
      statusBar.setText("Open URL canceled");
    return;
  }
    

  public void processSaveFile() {
  
    PrintStream outStream;
    
    if (SHOE_File == null) {
      /* IE doesn't like like applets that create file dialogs. Therefore we
	 don't create the dialog until we need it */
      if (saveFileDialog == null)
	saveFileDialog = new FileDialog(this, "Save File", FileDialog.SAVE);
      saveFileDialog.show();
      if (saveFileDialog.getFile() != null)
	SHOE_File = new File(saveFileDialog.getDirectory(), saveFileDialog.getFile());
    }
    if (SHOE_File != null) {
      statusBar.setText("Saving File: "+SHOE_File);

      try {
	outStream = new PrintStream(new FileOutputStream(SHOE_File));
      } catch (IOException ex) {
	//	notifyUser("Error opening file " + SHOE_File);
      statusBar.setText("Error opening File: "+SHOE_File);

	return;
      }
      doc.printHTML(outStream);
      outStream.close();

	//     notifyUser("Saved file " + SHOE_File);
      statusBar.setText("Saved File: "+SHOE_File);

    }
  }

  public void processSaveAsFile() {

    PrintStream outStream;

    /* IE doesn't like like applets that create file dialogs. Therefore we
       don't create the dialog until we need it */
    if (saveFileDialog == null)
      saveFileDialog = new FileDialog(this, "Save File", FileDialog.SAVE);    
    saveFileDialog.show();
    if (saveFileDialog.getFile() != null) {
      SHOE_File = new File(saveFileDialog.getDirectory(), saveFileDialog.getFile());
      statusBar.setText("Saving File: "+SHOE_File);

      try {
	outStream = new PrintStream(new FileOutputStream(SHOE_File));
      } catch (IOException ex) {
	//	notifyUser("Error opening file " + SHOE_File);
      statusBar.setText("Error opening File: "+SHOE_File);

	return;
      }

      //    System.out.println(doc);
      //    System.out.println(outStream);
      doc.printHTML(outStream);
      outStream.close();

	//     notifyUser("Saved file " + SHOE_File);
      statusBar.setText("Saved File: "+SHOE_File);

    }
  }

  /** Perform the Validate action. Re-parses the current document and
   displays any errors that it may have. */
  public void processValidate() {

    /* if in the Source view, we must reparse the
       string displayed by the text component. */
    if (currentView == V_SOURCE) {
      loadPageFromString(sourceText.getText());
      sourceEdit = false;
    }
    else {                   // we are in another view
      /* as a hack, we avoid problems with validating a document twice
	 and ensure that all error messages are properly
	 displayed by getting the string that represents the document
	 and re-parsing that string. */
      ByteArrayOutputStream arrayStream;
      PrintStream htmlStream;
      
      arrayStream = new ByteArrayOutputStream();
      htmlStream = new PrintStream(arrayStream);
      doc.printHTML(htmlStream);
      loadPageFromString(arrayStream.toString());
    }
  }

  public void processCopy() {
    notifyUser("Copy");
  }

  public void processPaste() {
    notifyUser("Paste");
  }

  public void processViewClaimPhrases() {

    Instance inst;
    Enumeration enum;

    // reparse if edits were made in the Source view
    if (currentView == V_SOURCE && sourceEdit == true) {
      loadPageFromString(sourceText.getText());
      sourceEdit = false;
    }

    currentView = V_PHRASE;
    ((CardLayout)viewPanel.getLayout()).show(viewPanel, "SHOE");
    inst = (Instance)instList.getSelectedObject();
    if (inst != null) {
      enum = inst.getDeclarations();
      replaceClaimList(enum);
    }
  }

  public void processViewClaimPredicates() {

    Instance inst;
    Enumeration enum;

    // reparse if edits were made in the Source view
    if (currentView == V_SOURCE && sourceEdit == true) {
      loadPageFromString(sourceText.getText());
      sourceEdit = false;
    }

    currentView = V_PREDICATE;
    ((CardLayout)viewPanel.getLayout()).show(viewPanel, "SHOE");
    inst = (Instance)instList.getSelectedObject();
    if (inst != null) {
      enum = inst.getDeclarations();
      replaceClaimList(enum);
    }
  }

  public void processViewSource() {

    currentView = V_SOURCE;
    displayDocInSourceView(doc);
    ((CardLayout)viewPanel.getLayout()).show(viewPanel, "Source");
  }

  /* displays a single summary of the SHOE tags on the page */
  public void processViewSummary() {

    // reparse if edits were made in the Source view
    if (currentView == V_SOURCE && sourceEdit == true) {
      loadPageFromString(sourceText.getText());
      sourceEdit = false;
    }

    currentView = V_SUMMARY;
    displayDocInSummaryView(doc);
    ((CardLayout)viewPanel.getLayout()).show(viewPanel, "Source");
  }


  public void processAddInst() {

    Use_Ontology[] useOnts;

    InstWindow popup = new InstWindow(this, InstWindow.ADD, doc, instList,ontman, pageURL);
    // Utilities.showDialog(popup);
    popup.showWorkaround();
    // System.out.println("after showDialog()");

    Enumeration enum = doc.getPSHOE();
    replaceInstList(enum);
   
    // int size=instList.getItemCount();
    int size=instList.countItems();
    if(size>0) {
      instList.select(size-1);	
      enableInstOps();
      Instance inst = (Instance)instList.getSelectedObject();

    /* we replace the claim list to show the name of the instance */
      enum = inst.getDeclarations();
      replaceClaimList(enum);
      useOnts = inst.getUseOnts();
      replaceUseOntList(useOnts);
    }
  }

  public void processEditInst() {
    Instance inst = (Instance)instList.getSelectedObject();
    int index = instList.getSelectedIndex();

    InstWindow popup = new InstWindow(this, InstWindow.EDIT, doc, instList, ontman, pageURL);
    // Utilities.showDialog(popup);
    popup.showWorkaround();

    Enumeration enum = doc.getPSHOE();
    replaceInstList(enum);            // do we have to replace the whole list?
    instList.select(index);	

    /* we replace the claim list in case the name of the instance changed */
    enum = inst.getDeclarations();
    replaceClaimList(enum);         
  }

  public void processDelInst() {

    Instance inst = (Instance)instList.getSelectedObject();
    int index = instList.getSelectedIndex();

    instList.delItem(index);
    doc.deleteSHOEObject(inst);
    useOntList.clear();
    claimList.clear();
    deselectInst();
  }

  public void processAddUseOnt() {

    Use_Ontology[] useOnts;

    Instance inst = (Instance)instList.getSelectedObject();
    UseOntWindow popup = new UseOntWindow(this, UseOntWindow.ADD, ontman,
					  inst, useOntList, instList);
    // Utilities.showDialog(popup);
    popup.showWorkaround();

    useOnts = inst.getUseOnts();
    replaceUseOntList(useOnts);
  }

  public void processEditUseOnt() {

    Instance inst = (Instance)instList.getSelectedObject();
    UseOntWindow popup = new UseOntWindow(this, UseOntWindow.EDIT, ontman,
					  inst, useOntList, instList);
    // Utilities.showDialog(popup);
    popup.showWorkaround();
  }

  public void processDelUseOnt() {

    Instance inst = (Instance)instList.getSelectedObject();

    Use_Ontology uont = (Use_Ontology)useOntList.getSelectedObject();
    int index = useOntList.getSelectedIndex();

    useOntList.delItem(index);
    inst.deleteUseOnt(uont);
    disableUseOntOps();
  }

  public void processAddClaim() {

    Instance inst = (Instance)instList.getSelectedObject();
    /*    ClaimWindow popup = new ClaimWindow(this, ClaimWindow.ADD, 
					inst, claimList, ontman);
					*/
    ClaimDlg dialog = new ClaimDlg((Frame)this, ClaimDlg.ADD,
				   claimList, inst, ontman);
    // Utilities.showDialog(dialog);   
    dialog.showWorkaround();
    
    // Enumeration enum = inst.getDeclarations();
    // replaceClaimList(enum);
    
  }

  public void processEditClaim() {

    Instance inst = (Instance)instList.getSelectedObject();
    /*
    ClaimWindow popup = new ClaimWindow(this, ClaimWindow.EDIT, 
					inst, claimList, ontman);
  */
    ClaimDlg dialog = new ClaimDlg((Frame)this, ClaimDlg.EDIT, 
				   claimList, inst, ontman);
    // Utilities.showDialog(dialog);   
    dialog.showWorkaround();
    
    // Enumeration enum = inst.getDeclarations();
    // replaceClaimList(enum);

  }

  public void processDelClaim() {

    Instance inst = (Instance)instList.getSelectedObject();

    Declaration dec = (Declaration)claimList.getSelectedObject();
    int index = claimList.getSelectedIndex();

    /* don't allow the deletion of an unnamed subject */
    if (dec != null) {
      claimList.delItem(index);
      inst.deleteDeclaration(dec);
      disableClaimOps();
      Enumeration enum = inst.getDeclarations();
      replaceClaimList(enum);
    }
  }

  public void deselectInst() {

    if (instList.getSelectedIndex() >= 0) 
      instList.deselect(instList.getSelectedIndex());
    disableClaimOps();
    disableUseOntOps();
    disableInstOps();
    useOntList.clear();
    claimList.clear();
  }

  /** Enable those menus that require a document to be loaded. */
  public void enableDocMenus() {

    if (ApplType.getIsApplication()) {
      menuBar.getMenu("File").getItem("Save File").enable();
      menuBar.getMenu("File").getItem("Save File As").enable();
      menuBar.getMenu("Bookmarks").getItem("Add Bookmark").enable();
    }
    menuBar.getMenu("Instance").enable();
    menuBar.getMenu("Instance").getItem("Add").enable();
    b_AddInst.enable();
    repaint();
    // validate();
  }

  public void enableInstOps() {

    menuBar.getMenu("Instance").getItem("Edit").enable();
    menuBar.getMenu("Instance").getItem("Delete").enable();
    menuBar.getMenu("Use Ontology").enable();
    menuBar.getMenu("Claim").enable();
    b_EditInst.enable();
    b_DelInst.enable();
    b_AddUseOnt.enable();
    b_AddClaim.enable();
    repaint();
    // validate();
  }

  public void enableClaimOps() {
    menuBar.getMenu("Claim").getItem("Edit").enable();
    menuBar.getMenu("Claim").getItem("Delete").enable();
    b_EditClaim.enable();
    b_DelClaim.enable();
    repaint();
    // validate();
  }

  public void enableUseOntOps() {
    menuBar.getMenu("Use Ontology").getItem("Edit").enable();
    menuBar.getMenu("Use Ontology").getItem("Delete").enable();
    b_EditUseOnt.enable();
    b_DelUseOnt.enable();
    repaint();
    // validate();
  }

  public void disableInstOps() {

    menuBar.getMenu("Instance").getItem("Edit").disable();
    menuBar.getMenu("Instance").getItem("Delete").disable();
    menuBar.getMenu("Use Ontology").disable();
    menuBar.getMenu("Claim").disable();
    b_EditInst.disable();
    b_DelInst.disable();
    b_AddUseOnt.disable();
    b_AddClaim.disable();
  }

  public void disableClaimOps() {
    menuBar.getMenu("Claim").getItem("Edit").disable();
    menuBar.getMenu("Claim").getItem("Delete").disable();
    b_EditClaim.disable();
    b_DelClaim.disable();
  }

  public void disableUseOntOps() {
    menuBar.getMenu("Use Ontology").getItem("Edit").disable();
    menuBar.getMenu("Use Ontology").getItem("Delete").disable();
    b_EditUseOnt.disable();
    b_DelUseOnt.disable();
  }

  /** This routine updates the Instances list by clearing it and
     then adding an item for each instance. If the instance has a
     "name" relation, then the name of the instance is displayed as
     well. */
  public void replaceInstList(Enumeration enum) {

    SHOE_Obj obj;
    Instance inst;
    Enumeration claim_enum;
    String str;
    boolean bNamed;
    RelArg arg1, arg2;
    
    instList.clear();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Instance) {
        inst = (Instance)obj;
	bNamed=false;

	/* determine if the instance has a "name" relation */
	claim_enum= inst.getDeclarations();
	while (claim_enum.hasMoreElements()) {
	  obj = (SHOE_Obj)claim_enum.nextElement();
	  if(obj instanceof Relation) {
	    Relation rel = (Relation)obj;
	    if (rel.getBaseName().equals("name")) { 
	      arg1 = rel.getArg(1);
	      arg2 = rel.getArg(2);
	      if (arg1 != null && arg2 != null) {    // if both args present
		if (arg1.getValue().equals(inst.getKey())) {
		  str = arg2.getValue() + "  @  " + inst.getKey();
		  instList.addItem(str, inst);	
		  bNamed = true;
		  break;
		}
	      }
	    }
	  }
	}

	if(bNamed==false) 
	  instList.addItem(inst.getKey(), inst);
	
      }
    }
  }


  /*
    // No longer used. Was called by old replaceClaimList
  public void replaceInst() {
    SHOE_Obj obj;
    Instance inst;
    Enumeration claim_enum;
    String str;
    String arg;
    boolean bNamed;
    int index;
   
    inst=(Instance)instList.getSelectedObject();
    index = instList.getSelectedIndex();

    claim_enum= inst.getDeclarations();
    bNamed=false;
    while (claim_enum.hasMoreElements()) {
      obj = (SHOE_Obj)claim_enum.nextElement();
      if(obj instanceof Relation) {
	Relation rel = (Relation)obj;
	arg=rel.getArg(1).getValue();
	if(arg.equals(inst.getKey()) && rel.getBaseName().equals("name")) { 
	  str=(String)rel.getArg(2).getValue() + "  @  " + inst.getKey();
	  instList.replaceItem(str, inst,index);	
	  bNamed=true;
	  break;
	}
      }
    }
    if(bNamed==false) 
      instList.replaceItem(inst.getKey(), inst,index);	
  }
  */

  public void replaceUseOntList(Use_Ontology[] useOnts) {

    useOntList.clear();
    for (int i=0; i < useOnts.length; i++) {
      useOntList.addItem(useOnts[i].getOntLabel(), useOnts[i]);
    }
  }

  public void replaceClaimList(Enumeration enum) {

    SHOE_Obj obj;
    Declaration dec;
    Relation rel;
    String str;
    String key;
    Hashtable subjects;
    Vector subjClaims;         // vector of claims for a subject
    Enumeration subs;          // enumeration of subject claim vectors
    Enumeration claimEnum;     // enumeration of claims for a subject
    boolean named;
    boolean first;

    /* group claims together by their subjects */
    subjects = new Hashtable();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Declaration) {
	if (obj instanceof Relation) {
	  RelArg arg = ((Relation)obj).getArg(1);
	  if (arg != null)
	    key = arg.getValue();
	  else
	    key = "Error";
	}
	else         // if (obj instanceof Category)
	  key = ((Category)obj).getForText();

	subjClaims = (Vector)subjects.get(key);
	if (subjClaims == null) {           // no claims for this subject yet
	  subjClaims = new Vector();
	  subjClaims.addElement(obj);
	  subjects.put(key, subjClaims);
	}
	/* here we assume that updates to the object retrieved from the 
	   hashtable will affect the version stored in the hashtable */
	else {
	  /* if its a name relation, place at front of list, otherwise
	     add to end. Note getBaseName() may return a null value
	     (if the Relation is in error). Thus we issue the equals()
	     method on the "name" string */
	  if (obj instanceof Relation && 
	      "name".equals(((Relation)obj).getBaseName()))
	    subjClaims.insertElementAt(obj, 0);
	  else
	    subjClaims.addElement(obj);
	}
      }
    }

    /* add the results to the list */
    claimList.clear();
    subs = subjects.elements();
    while (subs.hasMoreElements()) {
      subjClaims = (Vector)subs.nextElement();
      claimEnum = subjClaims.elements();
      named = false;
      first = true;
      while (claimEnum.hasMoreElements()) {
	dec = (Declaration)claimEnum.nextElement();
	if (first == true) {
	  /* get key of the subject */
	  if (dec instanceof Relation) {
	    RelArg arg = ((Relation)dec).getArg(1);
	    if (arg != null)
	      key = arg.getValue();
	    else
	      key = "Error";
	  }
	  else  //if(obj instanceof Category)
	    key = ((Category)dec).getForText();

	  first = false;
	  if (dec instanceof Relation) {
	    rel = (Relation)dec;
	    /* if a name relation, display name with the key */
	    String relName = rel.getBaseName();
	    if (relName != null && relName.equals("name")) {
	      str = (String)rel.getArg(2).getValue();
	      str=str + "  (key = \"" + key + "\")";
	      claimList.addItem(str,rel);
	      named=true;
	      continue;            // do next iteration of the while loop
	    }
	  }
	  /* if no name relation, just display the key */
	  str="(key = \"" + key + "\")";
	  claimList.addItem(str,null);	  
	}
	/* display claims in normal format */
	if (currentView == V_PHRASE)
	  str = "     " + dec.getTextBySubject(ontman);
	else
	  str = "     " + dec.getPredicateText();
	claimList.addItem(str, dec);	
      }
    }
  }


  /* *** old one *****
 public void replaceClaimList(Enumeration enum) {

    SHOE_Obj obj;
    Declaration dec;
    
    claimList.clear();
    while (enum.hasMoreElements()) {
      obj = (SHOE_Obj)enum.nextElement();
      if (obj instanceof Declaration) {
        dec = (Declaration)obj;
        claimList.addItem(dec.get...(), dec);
      }
    }
  }
  */
  //////////////////////////////////////////////////////

  ///////// ClipBoard ---- 5/29/98, added by Qin ///////
  public void enableClaimPaste() {
    menuBar.getMenu("Claim").getItem("Paste").enable();
    menuBar.getMenu("Claim").getItem("Clear").enable();
  }  

  public void disableClaimPaste() {
    menuBar.getMenu("Claim").getItem("Paste").disable();
    menuBar.getMenu("Claim").getItem("Clear").disable();
  } 
 

  public void enableUseOntPaste() {
    menuBar.getMenu("Use Ontology").getItem("Paste").enable();
    menuBar.getMenu("Use Ontology").getItem("Clear").enable();
  }
  
  public void disableUseOntPaste() {
    menuBar.getMenu("Use Ontology").getItem("Paste").disable();
    menuBar.getMenu("Use Ontology").getItem("Clear").disable();
  }

  public void enableInstancePaste() {
    menuBar.getMenu("Instance").getItem("Paste").enable();
    menuBar.getMenu("Instance").getItem("Clear").enable();
  }
  
  public void disableInstancePaste() {
    menuBar.getMenu("Instance").getItem("Paste").disable();
    menuBar.getMenu("Instance").getItem("Clear").disable();
  }


  public Object cloneClaim(Object obj){    
    RelArg argList[];
    Declaration claim=(Declaration) obj;

    if (claim instanceof Relation) {
      Relation inrel = (Relation)claim;
      argList = inrel.getArgs();
      
      Relation rel = new Relation(inrel.getName());
      for (int i=0; i < argList.length; i++) 
	rel.addArg(new RelArg(i+1,argList[i].getValue()));
      return rel;
    }
    else if (claim instanceof  Category) {
      Category incat = (Category)claim;  
      Category cat = new Category(incat.getName(),incat.getForText());
      return cat;
    }
    
    return null;
  }


  public void clipClaimCopy(){
    Object obj=claimList.getSelectedObject();
    if(obj==null) return;
    clipClaimVector.addElement(cloneClaim(obj));
    enableClaimPaste();
  }
  
  public void clipClaimPaste(){ 
    Declaration obj;
    int size=clipClaimVector.size();
    Instance inst = (Instance)instList.getSelectedObject();

    for(int i=0;i<size;i++){
      obj=(Declaration)cloneClaim(clipClaimVector.elementAt(i)); 
      inst.addDeclaration(obj, ontman);
    
	/*
      if(obj instanceof Relation){
	claimList.addItem(((Relation)obj).getPredicateText(),obj); 
   
      }
      else if(obj instanceof Category) {
	claimList.addItem(((Category)obj).getPredicateText(),obj);
      }
	*/
    }
    Enumeration enum = inst.getDeclarations();
    replaceClaimList(enum); 

  }
  
  public void clipClaimClear(){
    clipClaimVector.removeAllElements();
    disableClaimPaste();
  }



  public Use_Ontology cloneUseOnt(Use_Ontology inont){   
    Use_Ontology ont = new Use_Ontology(inont.getId(), inont.getVersion(),
					inont.getPrefix(),inont.getUrl());
    return ont;
  }


  public void clipUseOntCopy(){
    Object obj=useOntList.getSelectedObject();
    if(obj==null) return;
    clipUseOntVector.addElement(cloneUseOnt((Use_Ontology)obj));
    enableUseOntPaste();
  }

  public void clipUseOntPaste(){
    Use_Ontology ont;
    int size=clipUseOntVector.size();
    Instance inst = (Instance)instList.getSelectedObject();

    for(int i=0;i<size;i++){
      ont=cloneUseOnt((Use_Ontology)clipUseOntVector.elementAt(i)); 
      try{
	inst.addUseOnt(ont);        
	useOntList.addItem(ont.getOntLabel(),ont);     
      } catch (DupeKeyException ex) {
	notifyUser("Prefix is already used. New record not added");
      }
    }
  }
  
  public void clipUseOntClear(){
    clipUseOntVector.removeAllElements();
    disableUseOntPaste();
  }


  /* This should really be a method in Instance */
  public Instance cloneInstance(Instance ininst){   

    Use_Ontology[] useOnts;
    Instance inst = new Instance(ininst.getKey(), ininst.getDelegates());

    // must add Use Ontologies before Claims to avoid error messages 
    useOnts = ininst.getUseOnts();
    for (int i=0; i < useOnts.length; i++) {
      try{
	inst.addUseOnt(cloneUseOnt(useOnts[i]));       
      }catch (DupeKeyException ex) {
	//notifyUser("Prefix is already used. New record not added");
      }
    }

    for (Enumeration e = ininst.getDeclarations(); e.hasMoreElements() ;) 
      inst.addDeclaration((SHOE_Obj)cloneClaim(e.nextElement()),ontman);
	
    return inst;
  }


  public void clipInstanceCopy(){
    Object obj=instList.getSelectedObject();
    if(obj==null) return;
    clipInstanceVector.addElement(cloneInstance((Instance)obj));
    enableInstancePaste();
  }

  public void clipInstancePaste(){
    Instance inst;
    int size=clipInstanceVector.size();

    for(int i=0;i<size;i++){
      inst=cloneInstance((Instance)clipInstanceVector.elementAt(i));     
      instList.addItem(inst.getKey(), inst);     
      doc.addInstance(inst);
    }
  }
  
  public void clipInstanceClear(){
    clipInstanceVector.removeAllElements();
    disableInstancePaste();
  }



  /** Determines whether or not the Checkbox work-around must be used.
    See "Note: CheckboxMenuItem Events" in the action() method. */
  public void setCheckboxWorkAroundStatus() {

    if ((System.getProperty("os.name").equals("SunOS") ||
	 System.getProperty("os.name").equals("Solaris")) &&
	System.getProperty("java.vendor").startsWith("Sun")) {
      useCheckboxWorkAround = true;
      System.out.println("Using Checkbox Work-around");
    }
    else
      useCheckboxWorkAround = false;
  }


  public void notifyUser(String s) {

    Notification note;

    note = new Notification(this, "Error", s);
    note.showWorkaround();
  }

}

/* Threads used to implement the various actions in the event handler */

class OpenURLThread extends Thread {

  KA_App ka;

  public OpenURLThread(KA_App ka) {

    super("Open URL");
    this.ka = ka;
  }

  public void run() {

    ka.processOpenURL();
  }
}

class AddInstanceThread extends Thread {

  KA_App ka;

  public AddInstanceThread(KA_App ka) {

    super("Add_Instance");
    this.ka = ka;
  }

  public void run() {

    // System.out.println("AddInstanceThread started");
    ka.processAddInst();
  }
}

class EditInstanceThread extends Thread {

  KA_App ka;

  public EditInstanceThread(KA_App ka) {

    super("Edit Instance");
    this.ka = ka;
  }

  public void run() {

    ka.processEditInst();
  }
}

class AddUseOntThread extends Thread {

  KA_App ka;

  public AddUseOntThread(KA_App ka) {

    super("Add Use Ontology");
    this.ka = ka;
  }

  public void run() {

    ka.processAddUseOnt();
  }
}

class EditUseOntThread extends Thread {

  KA_App ka;

  public EditUseOntThread(KA_App ka) {

    super("Edit Use Ontology");
    this.ka = ka;
  }

  public void run() {

    ka.processEditUseOnt();
  }
}

class AddClaimThread extends Thread {

  KA_App ka;

  public AddClaimThread(KA_App ka) {

    super("Add Claim");
    this.ka = ka;
  }

  public void run() {

    ka.processAddClaim();
  }
}

class EditClaimThread extends Thread {

  KA_App ka;

  public EditClaimThread(KA_App ka) {

    super("Edit Claim");
    this.ka = ka;
  }

  public void run() {

    ka.processEditClaim();
  }
}
