package UserInt;

import java.awt.*;

/** This class creates a panel with scrollbars. It is a panel that includes
   two scrollbars and an interior panel on which all other components are
   drawn. */
public class MyScrollPane extends Panel {

  public static final int LINE_INC = 20;
  public static final int PAGE_INC = 80;
  public static final int BORDER_X = 0;
  public static final int BORDER_Y = 0;

  private Panel viewport;
  private Scrollbar hbar;
  private Scrollbar vbar;
  private Panel interior;

  /* coordinates of visible upper left corner. Changed by the scroll bar */
  private int tx = 0;
  private int ty = 0;

  private boolean resized = false;

  private Graphics intg;       // graphics object of interior panel
  /* private Component gp;                        // grand-parent
   */

  public MyScrollPane(Panel interior) {

    this.interior = interior;
    viewport = new Panel();
    viewport.add(interior);
    viewport.validate();
    viewport.show();
    // viewport.setLayout(new BorderLayout());
    hbar = new Scrollbar(Scrollbar.HORIZONTAL, 0, 1, 0, 1);
    vbar = new Scrollbar(Scrollbar.VERTICAL, 0, 1, 0, 1);

    setLayout(new BorderLayout());
    add("Center", viewport);
    add("South", hbar);
    add("East", vbar);

    hbar.setLineIncrement(LINE_INC);
    hbar.setPageIncrement(PAGE_INC);
    vbar.setLineIncrement(LINE_INC);
    vbar.setPageIncrement(PAGE_INC);

    System.out.println("Initializing increments");
    System.out.println("HLine = " + hbar.getLineIncrement() +
		       ", VLine = " + vbar.getLineIncrement());
    // validate();
    // resizeScrollbars(); 
    // gp = this.getParent().getParent();
  }


  public void scrollTo(int x, int y) {

    System.out.println("scrollTo(" + x + "," + y + ")");
    interior.move(-x, -y);
    interior.repaint();

    /* move scroll bars to match current position of frame */
    /*
    if (hbar.getValue() != x)
      hbar.setValue(x);
    if (vbar.getValue() != y)
      vbar.setValue(y);
      */
    System.out.println("Increments after scrollTo:");
    System.out.println("HLine = " + hbar.getLineIncrement() +
		       ", VLine = " + vbar.getLineIncrement());
  }

  public void paint(Graphics g) {

    System.out.println("Called paint()");

    scrollTo(tx, ty);
    interior.repaint();

    /*
    Point p = interior.location();
    System.out.println("Interior origin = (" + p.x + "," + p.y + ")");
    */

    /* in Java 1.0, Containers do not have a paint method */
    /*
    try {
      super.paint(g);                   // Java 1.1
    } catch (NoSuchMethodError ex) {
      ((Component)this).paint(g);                       // Java 1.0
    }
    */
  }

  public void paintComponents(Graphics g) {

    System.out.println("Called paintComponents()");
    super.paintComponents(g);
  }

  public void layout() {

    System.out.println("Called layout()");
    super.layout();
  }

  public void validate() {

    System.out.println("Called validate()");
    super.validate();
  }


  /** Get the height of the viewport */
  public final int getVisibleHeight() {

    return viewport.size().height;
  }

  /** Get the width of the viewport */
  public final int getVisibleWidth() {

    return viewport.size().width;
  }

  /* This is automatically called whenever the panel is resized */
  public void reshape(int x, int y, int width, int height) {

    super.reshape(x, y, width, height);
    if (width != size().width || height != size().height)
      resizeScrollbars();
  }

  /* This function should be called after adding objects to the canvas to
     make sure the scroll bars cover the entire area */
  public void resizeScrollbars() {

    int sMinY, sMaxY, sMinX, sMaxX;
    int vbar_val, hbar_val;

    int min_x = 0;
    int min_y = 0;
    int max_x = interior.preferredSize().width;
    int max_y = interior.preferredSize().height;
    int vslideSize = getVisibleHeight();
    int hslideSize = getVisibleWidth();

    if (vslideSize <= 0)
      vslideSize = 1;
    if (hslideSize <= 0)
      hslideSize = 1;
    if (max_x < hslideSize)
      max_x = hslideSize;
    if (max_y <= vslideSize)
      max_y = vslideSize;

    /* be sure not to change scrollbar position when resizing */
    vbar_val = vbar.getValue();
    hbar_val = hbar.getValue();
    vbar.setValues(vbar_val, vslideSize, min_y, max_y);
    hbar.setValues(hbar_val, hslideSize, min_x, max_x);

    System.out.println("Increments after resizing:");
    System.out.println("HLine = " + hbar.getLineIncrement() +
		       ", VLine = " + vbar.getLineIncrement());

    System.out.println("Set vbar to " + vslideSize + "," + min_y + "," + max_y);
    System.out.println("Set hbar to " + hslideSize + "," + min_x + "," + max_x);

    /*
    vbar.setValues(vbar_val, getVisibleHeight(), sMinY, sMaxY);
    hbar.setValues(hbar_val, getVisibleWidth(), sMinX, sMaxX);
    */

    /* must reset scroll bar adjusment when changing the bar's value */
    /*
    tx = 0;
    ty = 0;
    */

    /*
    System.out.println("Min and max are (" + min_x + "," + min_y + ") and (" +
		       max_x + "," + max_y + ")");
    System.out.println("Set window size to (" + sMinX + "," + sMinY + ")(" +
		       sMaxX + "," + sMaxY + ")");
    resized = true;
		       */

  }


  public boolean handleEvent(Event evt) {
    switch (evt.id) {
    case Event.SCROLL_LINE_UP:
    case Event.SCROLL_LINE_DOWN:
    case Event.SCROLL_PAGE_UP:
    case Event.SCROLL_PAGE_DOWN:
    case Event.SCROLL_ABSOLUTE:
      if (evt.target == vbar) {
	ty = ((Integer)evt.arg).intValue();
	// interior.move(-tx,-ty);
	System.out.println("vbar: ty = " + ty);
	repaint();
      }
      if (evt.target == hbar){
	tx = ((Integer)evt.arg).intValue();
	// interior.move(-tx,-ty);
	System.out.println("hbars tx = " + tx);
	repaint();
      }
    }
    return super.handleEvent(evt);
  }


}



